////////////////////////////////////////////////////////////////////////////////
///
///		Title		: UI Framework Base Component File
///
///		Copyright (c) 2006, NHN Co., Ltd. ALL RIGHTS RESERVED
///
///		Team		: Desktop Service Development
///
////////////////////////////////////////////////////////////////////////////////
#pragma once

#include <atlwin.h>
#include <atltypes.h>
#include <atlcoll.h>

#include "Common.h"
#include "RTTI.h"
#include "MouseEvent.h"
#include "KeyEvent.h"
#include "PropertyEvent.h"
#include "ComponentUI.h"
#include "XMLUtil.h"
#include "Object.h"

static const int MAX_COMPONENT_NAME = 128;
///< virtual ProcessWindowMessage
#define VBEGIN_MSG_MAP(theClass) \
public: \
	virtual BOOL ProcessWindowMessage(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, LRESULT& lResult, DWORD dwMsgMapID = 0) \
	{ \
	BOOL bHandled = TRUE; \
	(hWnd); \
	(uMsg); \
	(wParam); \
	(lParam); \
	(lResult); \
	(bHandled); \
	switch(dwMsgMapID) \
		{ \
	case 0:
#define USERDEFINED_MESSAGE_HANDLER(func) \
	if(WM_USER < uMsg) \
	{ \
	bHandled = TRUE; \
	lResult = func(uMsg, wParam, lParam, bHandled); \
	if(bHandled) \
	return TRUE; \
	}

class NPCOMMONUI_API ComponentProperty
{
public:
	static const TCHAR* MouseEnter;
	static const TCHAR* MouseExit;
	static const TCHAR* OnCreate;
	static const TCHAR* OnDestroy;
};
class HorizonAlignment
{
public:
	enum eHorizonAlignment
	{
		Default,
		Left,
		Center,
		Right,
		Stretch
	};
};
class VerticalAlignment
{
public:
	enum eVerticalAlignment
	{
		Default,
		Top,
		Center,
		Bottom,
		Stretch
	};
};
/**
* @ingroup 	Base UI
* @brief 	{kr} UIless Base Component {/kr}
* @date		10/09/2006
* @author	telltale
* @remark	
*/
class CContainer;
class CRootContainer;
class Binding;
class NPCOMMONUI_API CComponent : public Object
{
public:
	CComponent(void);
	virtual ~CComponent(void);
	
	void Create();
	void Destroy();

	/**
	* @brief 	add child component
	* @param	pChild	child component
	*/
	virtual	void	Add(CComponent* pChild) {}
	/**
	* @brief 	remove child component
	* @param	pChild child component
	* @return	If the function succeeds, the return value is nonzero
	*/
	virtual BOOL	Remove(CComponent* pChild) { return FALSE; }
	/**
	* @brief 	remove all child component
	* @return	If the function succeeds, the return value is TRUE
	*/
	virtual BOOL	RemoveAll() { return FALSE; }
	/**
	* @brief 	modify position and dimension of the specified component
	* @param	nX	specifies the new position of the left side of the component
	* @param	nY	specifies the new position of the top side of the component
	* @param	nWidth specifies the new width of the component
	* @param	nHeight specifies the new height of the component
	* @param	bRepaint	specifies whether the component is to be repained.
	* @return	If the function succeeds, the return value is nonzero
	*/
	virtual BOOL	MoveWindow(int nX, int nY, int nWidth, int nHeight, BOOL bRepaint=FALSE);
	virtual BOOL	MoveWindow(LPRECT lpRect, BOOL bRepaint=FALSE);
	virtual BOOL SetActualBound(int nX, int nY, int nWidth, int nHeight);

	/**
	* @brief 	this function retrieves the coordinates of a window area
	* @param	lpRect	Pointer to a RECT that receives the client area
	* @return	If function succeeds, the return value is TRUE
	*/
	virtual BOOL GetClientRect(LPRECT lpRect);
	/**
	* @brief 	this function retrieves the coordinates of a window area. The dimensions are given in screen coordinates that are relative to the upper-left corner of the screen.
	* @param	lpRect	Pointer to a RECT that receives the client area
	* @return	If function succeeds, the return value is TRUE
	*/
	BOOL	GetWindowRect(LPRECT lpRect);
	/**
	* @brief 	this function retrieves the coordinates of a component area.
	* @param	lpRect	Pointer to a RECT that receives the client area
	* @return	If function succeeds, the return value is TRUE
	*/
	BOOL	GetBounds(LPRECT lpRect);
	BOOL	GetOriginalBounds(LPRECT lpRect);
	BOOL	GetDrawingBounds(LPRECT lpRect);
	/**
	* @brief 	this function adds a rectangle to the specified window's update region.
	* @param	pRect	Pointer to a RECT structure the contains the client coordinates of the rectangle to be added to update region
	* @param	bErase	specifies whether the background within the update region is to be a erased when the update region is processed.
	* @return	If function succeeds, the return value is TRUE
	*/
	BOOL	InvalidateRect(LPCRECT pRect, BOOL bErase=TRUE);
	BOOL Invalidate() { return InvalidateRect(NULL); }
	BOOL	InvalidateImmediately();
	/**
	* @brief 	this function enable window
	*/
	virtual BOOL EnableWindow(BOOL bEnabled);
	BOOL IsWindowEnabled();
	
	/**
	* @brief 	converts the client coordinates of a given point on the display to screen coordinates
	* @param	lpPoint	Pointer to a POINT that contains the client coordinates to be converted.
	* @return	If function succeeds, the return value is TRUE
	*/
	BOOL ClientToScreen(LPPOINT lpPoint);
	/**
	* @brief 	converts the client coordinates of a given rectangle on the display to screen coordinates
	* @param	lpPoint	Pointer to a RECT that contains the client coordinates to be converted.
	* @return	If function succeeds, the return value is TRUE
	*/
	BOOL ClientToScreen(LPRECT lpRect);
	/**
	* @brief 	converts the screen coordinates of a given point on the display to client coordinates.
	* @param	lpPoint Pointer to a POINT that contains the screen coordinates to be converted.
	* @return	If function succeeds, the return value is TRUE
	*/
	BOOL ScreenToClient(LPPOINT lpPoint);
	/**
	* @brief 	converts the screen coordinates of a given rectangle on the display to client coordinates.
	* @param	lpPoint Pointer to a RECT that contains the screen coordinates to be converted.
	* @return	If function succeeds, the return value is TRUE
	*/
	BOOL ScreenToClient(LPRECT lpRect);
	/**
	* @brief 	show or hide this component depending on the value of parameter bVisible
	* @param	bVisible	if TRUE, show this component otherwise hide this component
	* @return	none
	*/
	virtual void	SetVisible(BOOL bVisible);
	virtual void OnParentVisibleChanged(CComponent* parent, BOOL bVisible) {}
	/**
	* @brief 	whether this component should be visible when its parent is visible
	* @return	TRUE if this component is visible, FALSE otherwise
	*/
	virtual BOOL	IsVisible();
	
	/**
	 * \brief {kr} mouse over flag {/kr}
	 * \return TRUE if your mouse is over this component , FALSE otherwise
	 */
	BOOL IsMouseOver();
	/**
	* @brief 	{kr} mouse/key event ó   Ѵ. {/kr}
	*/
	void SetEventRoute(BOOL bEnable)
	{
		m_bEventRouteFlag = bEnable;
	}
	BOOL IsEventRoute() { return m_bEventRouteFlag; }
	/**
	* @brief 	set parent container
	* @param	Parent	parent container
	*/
	void	SetParent(CContainer* pParent);
	/**
	* @brief 	get parent container
	* @return	the return value is parent container
	*/
	CContainer*	GetParent();	
	/**
	* @brief 	a position check contains boundary of component
	* @param	pt	a position
	* @return	If the function succeeds, the return value is TRUE
	*/
	BOOL	PtInRect(POINT pt);
	virtual BOOL	HitTest(POINT pt);
	BOOL	ParentPtInRect(POINT pt);
	/**
	* @brief 	the SetFocus function sets the keyboard focus to the specified component.
	* @return	If the function succeeds, the return value is the handle to the component had the keyboard focus
	*/
	CComponent* SetFocus();
	/**
	* @brief 	the GetFocus fn retrieves the component that havs the keyboard focus.
	* @return	the return value is the component with the keyboard focus.
	*/
	CComponent* GetFocus();
	/**
	* @brief 	this function sets the mouse capture to the specified window belonging to the current thread
	* @return	none
	*/
	void	SetCapture();
	/**
	* @brief 	this function releases the mouse capture
	* @return	none
	*/
	void	ReleaseCapture();
	/**
	* @brief 	this function send the specified message to a component
	* @param	uMsg	specifies the message to be sent
	* @param	wParam	specifies additional message-specific information
	* @param	lParam	specifies additional message-specific information
	* @return	the return value specifies the result of the message processing
	*/
	LRESULT	SendMessage(UINT uMsg, WPARAM wParam, LPARAM lParam);
	/**
	* @brief 	this function send the specified message to a component
	* @param	uMsg	specifies the message to be sent
	* @param	wParam	specifies additional message-specific information
	* @param	lParam	specifies additional message-specific information
	* @return	the return value specifies the result of the message processing
	*/
	LRESULT PostMessage(UINT uMsg, WPARAM WPARAM, LPARAM lParam);
	/**
	* @brief 	set the component id
	* @param	pszcName	name, id
	* @return	
	*/
	void	SetName(LPCTSTR pszcName);
	///<  get component's name
	LPCTSTR	GetName() { return m_szName; }
	virtual CComponent* FindName(LPCTSTR pszcName);
	virtual CComponent* FindParentName(LPCTSTR pszcName);
	virtual LPCTSTR GetClassName() { return _T("Component"); }

	/**
	* @brief 	set the ui object that renders this component
	* @param	pComponentUI	the ComponentUI object
	* @return	none
	*/
	virtual void	SetUI(IComponentUI* pComponentUI);
	/**
	* @brief 	get the ui object that renders this component
	* @return	the ComponentUI object
	*/
	IComponentUI*	GetUI();
	/**
	* @brief 	this function retrieves handle to owner window
	* @return	handle to owner window
	*/
	HWND	GetOwnerWindow();
	/**
	* @brief 	add the specified mouse listener to receive mouse event from this component
	* @param	pMouseListener	mouse listener
	* @return	none
	*/
	void	AddMouseListener(IMouseListener* pMouseListener);
	/**
	* @brief 	remove the specified mouse listener to receive mouse event from this component
	* @param	pMouseListener	mouse listener
	* @return	none
	*/
	void	RemoveMouseListener(IMouseListener* pMouseListener);
	/**
	* @brief 	add the specified key listener to receive key event from this component
	* @param	pKeyListener	key listener
	* @return	none
	*/
	void	AddKeyListener(IKeyListener* pKeyListener);
	/**
	* @brief 	remove the specified key listener to receive key event from this component
	* @param	pKeyListener	key listener
	* @return	none
	*/
	void	RemoveKeyListener(IKeyListener* pKeyListener);
	/**
	* @brief 	add the specified listener to receive propertychange event from this component
	* @param	pPropertyListener propertychanged listener
	* @return	none
	*/
	void	AddPropertyChangedListener(IPropertyChangedListener* pPropertyChangedListener);
	/**
	* @brief 	add the specified listener to receive propertychange event from this component
	* @param	pPropertyListener propertychanged listener
	* @return	none
	*/
	void	RemovePropertyChangedListener(IPropertyChangedListener* pPropertyChangedListener);

	virtual Binding* CreateBinding();
	void SetBinding(LPCTSTR propertyName, Binding* binding);
	
	/**
	* @brief 	get root container
	* @return	If the function succeeds, the return is RootContainer
	*/
	CRootContainer*	GetRootContainer();
	/**
	* @brief 	get preferred size
	* @param	
	* @return	
	*/
	virtual BOOL GetPreferredSize(LPSIZE lpSize);
	virtual BOOL GetOrginPreferredSize(LPSIZE lpSize);

	HorizonAlignment::eHorizonAlignment GetHorizontalAligment() { return m_HorizontalAlignment; }
	void SetHorizontalAlignment(HorizonAlignment::eHorizonAlignment align) { m_HorizontalAlignment = align; }

	VerticalAlignment::eVerticalAlignment GetVerticalAlignment() { return m_VerticalAlignment; }
	void SetVerticalAlignment(VerticalAlignment::eVerticalAlignment align) { m_VerticalAlignment = align; }

	virtual void GetMargin(LPRECT margin);
	virtual void SetMargin(LPRECT margin);
	virtual void GetPadding(LPRECT padding) {}
	virtual void SetPadding(LPRECT padding) {}
	/**
	* @brief 	whenever create this component
	* @return	none
	*/
	virtual void OnCreate() {}
	/**
	* @brief 	whenever destroy this component
	* @return	none
	*/
	virtual void OnDestroy() {}
	/**
	* @brief 	draw component
	* @param	hDC	HDC from parent window OnPaint function
	*/
	virtual void OnDraw(HDC hDC) {}
	/**
	* @brief 	whenever change dimension
	* @param	nWidth	specifies the new width of the component
	* @param	nHeight	specifies the new height of the component
	*/
	virtual void OnSize(INT32 nWidth, INT32 nHeight) {}
	/**
	* @brief 	whenever changed parent dimension
	* @param	none
	* @return	none
	*/
	virtual void OnAncestorResized() {}
	/**
	* @brief 	when visible flag is TRUE
	*/
	virtual void OnShowNotify() {}
	/**
	* @brief 	when visible flag is FALSE
	*/
	virtual void OnHideNotify() {}
	/**
	* @brief 	when this component is added by parent container
	*/
	virtual void OnAddNotify() {}
	/**
	* @brief 	when this component is removed by parent container
	*/
	virtual void OnRemoveNotify() {}
	/**
	* @brief 	
	* @param	nMouseMsg	specifies the mouse message number
	* @return	Nonzero to halt further processing
	*/
	BOOL OnSetCursor(UINT nMouseMsg) 
	{ 
		::SetCursor(::LoadCursor(NULL, IDC_ARROW));
		return 0; 
	}
	/**
	* @brief 	the function is called immediately before losing the input focus
	* @param	pNewComponent	specifies a pointer to the component that receives the input focus
	* @return	none
	*/
	virtual void OnKillFocus(CComponent* pNewComponent) {}
	/**
	* @brief 	the function is called after gaining the input focus
	* @param	pOldComponent	the component loses the input focus
	* @return	none
	*/
	virtual void OnSetFocus(CComponent* pOldComponent) {}
	/**
	* @brief 	get absolute position of parent container
	* @param	pPt	absolute position point pointer
	* @return	If the function succeeds, the return is TRUE
	*/
	BOOL GetParentAbsolutePosition(LPPOINT pPt);
	/**
	* @brief 	pretranslate message
	* @param	pMsg		massage struct
	* @return	If the msaage is processed , return the TRUE or FALSE
	*/
	virtual BOOL PreTranslateMessage(MSG* pMsg) {return FALSE;}
	/**
	* @brief 	when the cursor leaves the client area of the window
	* @return	none
	*/
	virtual void OnMouseLeave();

	virtual void RecalcLayout(BOOL bRecursive=FALSE) {}
	virtual BOOL ModifyComponent(CContainer* pParent, TiXmlElement* pNode) { return FALSE; }
	virtual BOOL SetProperty(LPCTSTR property, CComVariant value) { return FALSE; }
	virtual BOOL GetProperty(LPCTSTR property, CComVariant& value) { return FALSE; }

	virtual void _SetScrollPos(int xPos, int yPos) {}

	static inline LONG GetRealValue(LONG value, LONG defaultValue=0) { return (value==LONG_MAX ? defaultValue : value); }

	DECLARE_ROOT_RTTI;

protected:
	/**
	* @brief 	whether user double-clicks the left mouse button
	* @param	nFlags	Indicates whether various keys are down
	* @param	point	specifies the x-and y-coordinate of the cursor
	* @return	none
	*/
	virtual void OnLButtonDblClk(UINT nFlags, CPoint point);
	/**
	* @brief 	whether user presses the left mouse button
	* @param	nFlags	Indicates whether various keys are down
	* @param	point	specifies the x-and y-coordinate of the cursor
	* @return	none
	*/
	virtual void OnLButtonDown(UINT nFlags, CPoint point);
	/**
	* @brief 	whether user releases the left mouse button
	* @param	nFlags	Indicates whether various keys are down
	* @param	point	specifies the x-and y-coordinate of the cursor
	* @return	none
	*/
	virtual void OnLButtonUp(UINT nFlags, CPoint point);
	/**
	* @brief 	whether user double-clicks the right mouse button
	* @param	nFlags	Indicates whether various keys are down
	* @param	point	specifies the x-and y-coordinate of the cursor
	* @return	none
	*/
	virtual void OnRButtonDblClk(UINT nFlags, CPoint point);
	/**
	* @brief 	whether user presses the right mouse button
	* @param	nFlags	Indicates whether various keys are down
	* @param	point	specifies the x-and y-coordinate of the cursor
	* @return	none
	*/
	virtual void OnRButtonDown(UINT nFlags, CPoint point);
	/**
	* @brief 	whether user releases the right mouse button
	* @param	nFlags	Indicates whether various keys are down
	* @param	point	specifies the x-and y-coordinate of the cursor
	* @return	none
	*/
	virtual void OnRButtonUp(UINT nFlags, CPoint point);
	/**
	* @brief 	whether the mouse cursor moves.
	* @param	nFlags	Indicates whether various keys are down
	* @param	point	specifies the x-and y-coordinate of the cursor
	* @return	none
	*/
	virtual void OnMouseMove(UINT nFlags, CPoint point);
	/**
	* @brief 	when the cursor enters the client area of the window
	* @return	none
	*/
	virtual void OnMouseEnter(CPoint point);
	/**
	* @brief 	whether the mouse cursor moves.
	* @param	nFlags	Indicates whether various keys are down
	* @param	nDelta	Indicates distance rotate
	* @param	point	specifies the x-and y-coordinate of the cursor
	* @return	none
	*/
	virtual void OnMouseWheel(UINT nFlags, short nDelta, CPoint point);
	/**
	* @brief 	when a nonsystem key is pressed
	* @param	nChar	specifies the virtual key code of the given key
	* @param	nRepCount	specifies the repeat count
	* @param	nFlags	specifies the scan code, key-transition code, previous key state, and context code
	* @return	none
	*/
	virtual void OnKeyDown(UINT nChar, UINT nRepCount, UINT nFlags);
	/**
	* @brief 	when a nonsystem key is released
	* @param	nChar	specifies the virtual key code of the given key
	* @param	nRepCount	specifies the repeat count
	* @param	nFlags	specifies the scan code, key-transition code, previous key state, and context code
	* @return	none
	*/
	virtual void OnKeyUp(UINT nChar, UINT nRepCount, UINT nFlags);
	/**
	* @brief 	when a user-defined message
	* @param	uMsg	specifies the message to be sent
	* @param	wParam	specifies additional message-specific information
	* @param	lParam	specifies additional message-specific information
	* @return	the return value specifies the result of the message processing
	*/
	virtual LRESULT OnMessage(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled);

	//friend class CContainer;

	///< parent container
	CContainer*	m_pParent;
	///< mouse enter flag
	BOOL	m_bEnter;
	///< mouse listener list
	CAtlList<IMouseListener*>	m_vMouseListener;
	///< key listener list
	CAtlList<IKeyListener*>		m_vKeyListener;
	///< propertychange lister list
	CAtlList<IPropertyChangedListener*> m_vPropertyListener;
	// binding
	CAtlList<Binding*> m_vBinding;

	///< component ui class
	IComponentUI*	m_pComponentUI;
	///< visible flag
	BOOL	m_bVisible;
	///< enable flag
	BOOL	m_bEnabled;
	///< {kr}  mouse/key event ó    {/kr}
	BOOL	m_bEventRouteFlag;
	///< name, id
	TCHAR	m_szName[MAX_COMPONENT_NAME];
	///< GetDC clip region
	HRGN m_hGetDCRgn;
	/**
	 * \brief {kr} ũ ġ  ´. {/kr}
	 * \param ppt {kr} ũ ġ  {/kr}
	 * \return If the function succeeds, the return is TRUE
	 */
	virtual BOOL _GetScrollPos(LPPOINT ppt) { return FALSE; }
	
	BOOL _IsWindowAccessible() { return IsWindowEnabled() && IsVisible(); }
	BOOL IsNCComponent();
	BOOL IsCaptured();
	
	virtual void FirePropertyChangedEvent(CPropertyChangeEvent* pPropertyEvent);
	
private:
	///< relative position and dimension
	RECT	m_rcBound;
	RECT	m_rcActualBound;

	HorizonAlignment::eHorizonAlignment m_HorizontalAlignment;
	VerticalAlignment::eVerticalAlignment m_VerticalAlignment;
	RECT m_Margin;

	void _DestroyBinding();

	void _RemoveParent();
	void _Monitor_CursorEnterExit(LPRECT prcBoundOld, LPRECT prcBoundNew);

VBEGIN_MSG_MAP(CComponent)	
	if (m_bEventRouteFlag)
	{
		USERDEFINED_MESSAGE_HANDLER(OnUserDefinedMessage)

		if (_IsWindowAccessible() == FALSE)
		{
			bHandled = FALSE;
			return 0;
		}		
		MESSAGE_HANDLER(WM_LBUTTONDBLCLK, OnLButtonDblClk)
		MESSAGE_HANDLER(WM_LBUTTONDOWN, OnLButtonDown)
		MESSAGE_HANDLER(WM_LBUTTONUP, OnLButtonUp)
		MESSAGE_HANDLER(WM_RBUTTONDBLCLK, OnRButtonDblClk)
		MESSAGE_HANDLER(WM_RBUTTONDOWN, OnRButtonDown)
		MESSAGE_HANDLER(WM_RBUTTONUP, OnRButtonUp)
		MESSAGE_HANDLER(WM_MOUSEMOVE, OnMouseMove)
		MESSAGE_HANDLER(WM_MOUSEWHEEL, OnMouseWheel)
		MESSAGE_HANDLER(WM_KEYDOWN, OnKeyDown)
		MESSAGE_HANDLER(WM_KEYUP, OnKeyUp)
		//MESSAGE_HANDLER(WM_MOUSELEAVE, OnMouseLeave)
		MESSAGE_HANDLER(WM_SETCURSOR, OnSetCursor)
	}
END_MSG_MAP()

	DECLARE_MESSAGE_HANDLER(OnUserDefinedMessage);
	DECLARE_MESSAGE_HANDLER(OnLButtonDblClk);
	DECLARE_MESSAGE_HANDLER(OnLButtonDown);
	DECLARE_MESSAGE_HANDLER(OnLButtonUp);
	DECLARE_MESSAGE_HANDLER(OnRButtonDblClk);
	DECLARE_MESSAGE_HANDLER(OnRButtonDown);
	DECLARE_MESSAGE_HANDLER(OnRButtonUp);
	DECLARE_MESSAGE_HANDLER(OnMouseMove);
	DECLARE_MESSAGE_HANDLER(OnMouseWheel);
	DECLARE_MESSAGE_HANDLER(OnKeyDown);
	DECLARE_MESSAGE_HANDLER(OnKeyUp);
	DECLARE_MESSAGE_HANDLER(OnSetCursor);
	//DECLARE_MESSAGE_HANDLER(OnMouseLeave);
};
